;;; DO NOT MODIFY THIS FILE
(if (featurep 'xlib-autoloads) (error "Already loaded"))

;;;### (autoloads nil "_pkg" "lisp/_pkg.el")

(package-provide 'xlib :version 1.14 :author-version "lg@xwem.org--2005/xlib--main--2.1--version-0" :type 'regular)

;;;***

;;;### (autoloads (xt-check-xlib) "xlib-testing" "lisp/xlib-testing.el")

(autoload 'xt-check-xlib "xlib-testing" "\
Interactively check xlib." t nil)

;;;***

;;;### (autoloads nil "xlib-version" "lisp/xlib-version.el")

(defconst xlib-version "lg@xwem.org--2005/xlib--main--2.1--version-0")

;;;***

;;;### (autoloads (X-EventHandler-runall X-EventHandler-disable X-EventHandler-enable X-EventHandler-rem X-EventHandler-isset X-EventHandler-add) "xlib-xr" "lisp/xlib-xr.el")

(autoload 'X-EventHandler-add "xlib-xr" "\
To event handlers list EVHLIST add event HANDLER.

HANDLER is function which should accept three arguments - xdpy(X-Dpy),
xwin(X-Win) and xev(X-Event).  Only events with type that in
EVTYPES-LIST are passed to HANDLER. By default all events passed.
PRIORITY is place in events handler list, i.e. when HANDLER will be
called. Higher priorities runs first.

Return new list, use it like `(setq lst (X-EventHandler-add lst 'handler))'." nil nil)

(autoload 'X-EventHandler-isset "xlib-xr" "\
Examine EVHLIST and return X-EventHandler with HANDLER, PRIORITY and EVTYPES-LIST.
If you does not specify PRIORITY and EVTYPES-LIST, only matching with HANDLER occurs.
If event handler not found - nil will be returned." nil nil)

(autoload 'X-EventHandler-rem "xlib-xr" "\
From EVHLIST remove event HANDLER with PRIORITY and EVTYPES-LIST.
If you does not specify PRIORITY and EVTYPES-LIST, only matching with HANDLER occurs.
Return new list, use it like `(setq lst (X-EventHandler-rem lst 'handler))'." nil nil)

(autoload 'X-EventHandler-enable "xlib-xr" "\
In event handlers list EVHLIST mark HANDLER with PRIORITY and EVTYPES-LIST as active." nil nil)

(autoload 'X-EventHandler-disable "xlib-xr" "\
In event handlers list EVHLIST mark HANDLER with PRIORITY and EVTYPES-LIST as inactive." nil nil)

(autoload 'X-EventHandler-runall "xlib-xr" "\
Run all event handlers in EVHLIST on XEV.
Signal `X-Events-stop' to stop events processing." nil nil)

(defvar X-default-timeout 60 "\
This should be big enought, larger than any XEmacs blocking.")

;;;***

(provide 'xlib-autoloads)
